const express = require("express");
const { execSync, exec } = require('child_process');
const { promisify } = require('util');
const execAsync = promisify(exec);
const path = require('path');
const fs = require('fs');
const fsAsync = require('fs/promises');
const os = require('os');
const setupPath = '/etc/45drives/simple-setup-log.json';
const serverInfoPath = '/etc/45drives/server_info/server_info.json'

function getSetupInfo() {
    try {
        if (!fs.existsSync(setupPath)) return { setupComplete: false };

        const setupData = JSON.parse(fs.readFileSync(setupPath, 'utf8'));
        const serverInfo = JSON.parse(fs.readFileSync(serverInfoPath, 'utf8'));
        const localIP = getLocalIP();
        const hostname = os.hostname();

        let matchingEntries = [];

        for (const [entryIP, info] of Object.entries(setupData)) {
            if (!info || typeof info !== "object") continue;

            const entry = info;

            if (entryIP === localIP) {
                const matchesHostname = entry.serverName === hostname;
                matchingEntries.push({
                    ip: entryIP,
                    ...entry,
                    matchesHostname
                });
            }
        }

        if (matchingEntries.length === 0) {
            return { setupComplete: false };
        }

        // Sort: prefer hostname matches, then by most recent setupTime
        matchingEntries.sort((a, b) => {
            if (a.matchesHostname !== b.matchesHostname) {
                return b.matchesHostname - a.matchesHostname;
            }

            return new Date(b.setupTime).getTime() - new Date(a.setupTime).getTime();
        });

        const best = matchingEntries[0];

        return {
            setupComplete: true,
            serverName: best.serverName,
            shareName: best.shareName,
            setupTime: best.setupTime,
            moboMake: serverInfo["Motherboard"]["Manufacturer"],
            moboModel: serverInfo["Motherboard"]["Product Name"],
            serverModel: serverInfo["Model"],
            aliasStyle: serverInfo["Alias Style"],
            chassisSize: serverInfo["Chassis Size"],
        };

    } catch (error) {
        console.error("Error reading setup log:", error);
        return { setupComplete: false };
    }
}

function getLocalIP() {
    const interfaces = os.networkInterfaces();
    for (const iface of Object.values(interfaces)) {
        for (const config of iface) {
            if (config.family === "IPv4" && !config.internal) {
                return config.address;
            }
        }
    }
    return "127.0.0.1";
}

const app = express();
const port = 9095;

app.use(express.json()); // Middleware to parse JSON request bodies

// In-memory queue: Stores actions per IP
const actionQueue = {};

// ✅ Route to add a new action (specific to client IP)
app.post("/actions", (req, res) => {
    const clientIP = req.body.client_ip; // Get client IP address
    const action = req.body;

    // console.log("new Action:", action);

    if (!action) {
        return res.status(400).json({ error: "Action is required" });
    }

    if (!actionQueue[clientIP]) {
        actionQueue[clientIP] = [];
    }

    actionQueue[clientIP].push(action);
    res.json({ message: `Action added for ${clientIP}`, queueSize: actionQueue[clientIP].length });
});

// ✅ Route to get and remove the next available action (specific to client IP)
app.get("/actions", (req, res) => {
    const clientIP = req.query.client_ip;

    // console.log("client_ip:", clientIP);

    // console.log(actionQueue);

    if (!actionQueue[clientIP] || actionQueue[clientIP].length === 0) {
        return res.json({ message: "No actions available for your IP" });
    }

    const nextAction = actionQueue[clientIP].shift(); // Remove and return the first action for the client IP
    res.json(nextAction);
});

// ✅ Route to check queue size (specific to client IP)
app.get("/queue-size", (req, res) => {
    const clientIP = req.ip;
    const queueSize = actionQueue[clientIP] ? actionQueue[clientIP].length : 0;
    res.json({ queueSize });
});

async function findClientBackupPath() {
    try {
        const { stdout: zfsOut } = await execAsync('zfs list -H -o name,mountpoint');
        const zpools = zfsOut.split('\n')
            .map(line => line.trim().split(/\s+/))
            .filter(([name]) => name && !name.endsWith('-backup'));


        for (const [dataset, mountpoint] of zpools) {
            const clientBackupPath = path.join(mountpoint, 'client-backups');
            if (fs.existsSync(clientBackupPath)) {
                return clientBackupPath;
            }
        }
        return null;
    } catch (err) {
        console.error('Failed to find ZFS datasets:', err);
        return null;
    }
}

app.get('/backups', async (req, res) => {
    try {
        const backupRoot = await findClientBackupPath();
        const uuidDirs = await fsAsync.readdir(backupRoot);

        const results = [];

        for (const uuid of uuidDirs) {
            const backupDir = path.join(backupRoot, uuid);
            const entries = await walkDir(backupDir);
            entries.sort((a, b) => a.length - b.length);
            let baseFolder = uuid;
            let lastModified = 'Unknown';
            let keepSearchForBaseFolder = true;

            for (const entry of entries) {
                if (!await isDirectory(entry)) continue;

                // console.log(entry);
                
                // Recursively gather files to find last modified time
                const allFiles = await getAllFiles(entry);
                const fileCount = allFiles.filter(file => !file.isDirectory()).length;
                // console.log("fileCount", fileCount)
                if (allFiles.length) {
                    const times = await Promise.all(allFiles.map(async file => {
                        const stat = await fsAsync.stat(file.parentPath);
                        return stat.mtime;
                    }));
                    const mostRecent = new Date(Math.max(...times.map(t => t.getTime())));
                    lastModified = mostRecent.toISOString().split('T')[0]; // format YYYY-MM-DD
                }

                if (fileCount == 0 && keepSearchForBaseFolder) {
                    baseFolder = entry.replace(backupRoot, "").replace("/" + uuid + "/", "");
                }
                if (fileCount > 0 && keepSearchForBaseFolder) {
                    keepSearchForBaseFolder = false;
                    baseFolder = entry.replace(backupRoot, "").replace("/" + uuid + "/", "");
                }

            }

            const folders = baseFolder.split("/");
            let client = ""
            if (folders.length > 0) {
                baseFolder = baseFolder.replace(folders[0], "");
                client = folders[0];
            }

            results.push({
                uuid: uuid,
                folder: baseFolder,
                server: hostName,
                client: client,
                lastBackup: lastModified,
                onSystem: true,
            });
        }

        res.json(results);
    } catch (err) {
        console.error('Failed to list backups:', err);
        res.status(500).json({ error: 'Failed to list backups' });
    }
});

// Helper: recursively list all files
async function getAllFiles(dir) {
    const entries = await fsAsync.readdir(dir, { withFileTypes: true });

    return entries;
}

async function isDirectory(path) {
    try {
        const stats = await fsAsync.stat(path);
        return stats.isDirectory();
    } catch (err) {
        // Optional: handle error (e.g., path doesn't exist)
        return false;
    }
}

async function walkDir(dir, results = []) {
    const list = await fsAsync.readdir(dir, { withFileTypes: true });

    results.push(dir);

    for (const entry of list) {
        const fullPath = path.join(dir, entry.name);
        if (entry.isDirectory()) {
            await walkDir(fullPath, results);
        } else {
            results.push(fullPath);
        }
    }

    return results;
}

// ✅ Route to check queue size (specific to client IP)
app.get("/backups/files", async (req, res) => {
    const basePath = await findClientBackupPath();
    if (!basePath) return res.status(404).json({ error: 'Backup root not found' });

    const uuid = req.query.uuid;
    // console.log("uuid", uuid)
    // console.log("basePath", basePath)
    const folderPath = path.join(basePath, uuid);

    if (!fs.existsSync(folderPath)) return res.status(404).json({ error: 'Backup UUID not found' });

    // Recursively list all files
    const listFiles = (dir, relPath = '') => {
        const result = [];
        const entries = fs.readdirSync(dir);

        for (const entry of entries) {
            const fullPath = path.join(dir, entry);
            const rel = path.join(relPath, entry);
            const stat = fs.statSync(fullPath);

            if (stat.isDirectory()) {
                result.push(...listFiles(fullPath, rel));
            } else {
                result.push(rel);
            }
        }

        return result;
    };

    try {
        const files = listFiles(folderPath);
        res.json(files);
    } catch (err) {
        res.status(500).json({ error: 'Failed to list backup files' });
    }
});

app.get("/setup-status", (req, res) => {
    const defaultUsers = [
        "45drives", "_apt", "_rpc", "adm", "avahi", "backup", "bin", "ceph", "chrony", "clevis", "cockpit-ws", "cockpit-wsinstance",
        "daemon", "dbus", "dnsmasq", "ftp", "games", "gnats", "grafana", "halt", "irc", "libstoragemgmt", "list", "lp", "mail", "man",
        "messagebus", "news", "nobody", "operator", "pcp", "polkitd", "pollinate", "proxy", "root", "rpc", "rpcuser", "setroubleshoot",
        "shutdown", "statd", "sshd", "swtpm", "sync", "sys", "syslog", "systemd-coredump", "systemd-network", "systemd-resolve", "systemd-timesync",
        "tss", "tcpdump", "ubuntu", "unbound", "usbmux", "user", "www-data",

        // Newly added users
        "_chrony", "snapd", "systemd-oom",  // Debian/Ubuntu-specific
        "colord", "sssd", "nm-openvpn", "qemu", "firewalld", // RHEL/CentOS/Fedora
        "mysql", "postgres", // Database users
        "lxc", "docker", "libvirt-qemu", "libvirt-dnsmasq", // Virtualization & container users

        "usbmuxd", "rtkit", "pulse", "lightdm", "geoclue", "gluster", "saslauth", "libvirtdbus",
        'pipewire', 'flatpak', 'gdm'
    ];


    const { exec } = require('child_process');

    exec('cut -d: -f1 /etc/passwd', (err, stdout, stderr) => {
        if (err) {
            console.error('Error executing command:', err);
            res.json({ status: "error", err });
            return;
        }

        if (stderr) {
            console.error('stderr:', stderr);
            res.json({ status: "error", err: stderr });
            return;
        }

        // Split the output into an array by new lines
        const usersArray = stdout.split('\n').filter(user => user.length > 0);

        // Filter out default users
        const nonDefaultUsers = usersArray.filter(user => !defaultUsers.includes(user));

        // console.log('Array of users:');
        // console.log(nonDefaultUsers);

        res.json({ status: nonDefaultUsers.length > 0 ? "complete" : "not complete" });
    });

});

// ✅ Start the server
app.listen(port, () => {
    console.log(`Server is running on http://localhost:${port}`);
});

let bonjour = require('bonjour')();

const hostName = os.hostname();

// console.log(`Hostname: ${hostName}`);

// Function to start the server
function startBroadCasterServer() {
    console.log('Starting Houston Server...');
    bonjour = require('bonjour')();
    bonjour.find({ type: "houstonserver" }, async (service) => {
        if (service) {
            console.log(`Service ${service.name} is already registered.`);
        }
    });

    const setupInfo = getSetupInfo(); // Move this call inside the function call every time
    return bonjour.publish({
        name: hostName,
        type: 'houstonserver',
        protocol: 'tcp',
        port: port,
        host: hostName,
        txt: {
            info: 'Houston Server',
            setupComplete: setupInfo.setupComplete ? 'true' : 'false',
            serverName: setupInfo.serverName || '',
            shareName: setupInfo.shareName || '',
            setupTime: setupInfo.setupTime || '',
            moboMake: setupInfo.moboMake || '',
            moboModel: setupInfo.moboModel || '',
            serverModel: setupInfo.serverModel || '',
            aliasStyle: setupInfo.aliasStyle || '',
            chassisSize: setupInfo.chassisSize || ''
        },
    });
}

// Start the server initially
let service = startBroadCasterServer();

// Stop and restart the server every 5 seconds
setInterval(() => {
    console.log('Stopping Houston Server...');
    service.stop(() => {
        console.log('Houston Server stopped.');
        bonjour.destroy();
        service = startBroadCasterServer();
    });
}, 5000); // 5000ms = 5 seconds

// Gracefully clean up on exit
process.on('SIGINT', () => {
    console.log('Shutting down Houston Server...');
    service.stop(() => {
        bonjour.destroy();
        process.exit();
    });
});
